# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals, absolute_import, division

__author__ = """Stuart Littlefair"""
__email__ = 's.littlefair@shef.ac.uk'
__version__ = '1.1.2'


class DriverError(Exception):
    pass



from matplotlib import pyplot as plt


class Mimic(object):
    def __init__(self, height, width):
        dpi = 100
        figsize = (width/dpi, height/dpi)
        self.figure = plt.Figure(figsize=figsize, dpi=dpi)
        self.ax = self.figure.add_subplot(111)
        self.ax.set_axis_off()

    def update_mimic(self, telemetry):
        raise NotImplementedError



from __future__ import absolute_import, division, print_function

import pickle

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib import patches

import six
import numpy as np

from autobahn.wamp.exception import SerializationError

from hcam_widgets.tkutils import get_root
from hcam_widgets.widgets import IntegerEntry
from hcam_widgets.mimic import Mimic
from hcam_devices.models.slide import BLOCK_POS, UNBLOCK_POS
from hcam_widgets.misc import async_sleep

from twisted.internet.defer import inlineCallbacks

if not six.PY3:
    import Tkinter as tk
else:
    import tkinter as tk


SLEEP_BEFORE_MOVE = 0.5


class SlideFrame(tk.LabelFrame, Mimic):
    """
    Self-contained widget to deal with the focal plane slide
    """

    def __init__(self, master, show_mimic=True):
        """
        master  : containing widget
        """
        Mimic.__init__(self, height=200, width=100)
        tk.LabelFrame.__init__(self, master, text="Focal plane slide", padx=10, pady=4)

        # Top for table of buttons
        top = tk.Frame(self)

        # Define the buttons
        width = 8
        self.connect = tk.Button(
            top,
            fg="black",
            text="conn",
            width=width,
            command=lambda: self.action("connection.connect"),
        )

        self.home = tk.Button(
            top,
            fg="black",
            text="home",
            width=width,
            command=lambda: self.action("stage.home"),
        )

        self.block = tk.Button(
            top,
            fg="black",
            text="block",
            width=width,
            command=lambda: self.action("block"),
        )

        self.unblock = tk.Button(
            top,
            fg="black",
            text="unblock",
            width=width,
            command=lambda: self.action("unblock"),
        )

        self.gval = IntegerEntry(top, UNBLOCK_POS, None, True, width=4)

        self.goto = tk.Button(
            top,
            fg="black",
            text="goto",
            width=width,
            command=lambda: self.action("goto", self.gval.value()),
        )

        self.reset = tk.Button(
            top,
            fg="black",
            text="reset",
            width=width,
            command=lambda: self.action("reset"),
        )

        self.stop = tk.Button(
            top,
            fg="black",
            text="stop",
            width=width,
            command=lambda: self.action("stage.stop"),
        )

        self.enable = tk.Button(
            top,
            fg="black",
            text="enable",
            width=width,
            command=lambda: self.action("enable"),
        )

        self.disable = tk.Button(
            top,
            fg="black",
            text="disable",
            width=width,
            command=lambda: self.action("disable"),
        )

        self.restore = tk.Button(
            top,
            fg="black",
            text="restore",
            width=width,
            command=lambda: self.action("restore"),
        )

        # arrange the permanent ones
        self.connect.grid(row=0, column=0)
        self.home.grid(row=0, column=1)
        self.block.grid(row=0, column=2)
        self.unblock.grid(row=1, column=2)
        self.goto.grid(row=1, column=0)
        self.gval.grid(row=1, column=1)

        # set others according to expertlevel
        self.setExpertLevel()

        g = get_root(self).globals

        # widget for messages
        self.label = tk.Text(top, height=4, width=30, bg=g.COL["log"])
        self.label.configure(state=tk.NORMAL, font=g.ENTRY_FONT)
        self.label.grid(row=4, column=0, columnspan=3)

        # mimic
        self.showing_mimic = show_mimic
        if show_mimic:
            self.canvas = FigureCanvasTkAgg(self.figure, top)
            self.canvas.get_tk_widget().grid(row=5, column=0, columnspan=3)

        top.pack(pady=2)

    def update_mimic(self, telemetry):
        """
        Use current telemetry to update Mimic of the slide
        """
        if not telemetry or not self.showing_mimic:
            return

        try:
            pos = telemetry["position"]["current"].value
        except Exception:
            pos = telemetry["position"]["current"]

        state = telemetry["state"]
        connection_state = state["connection"]
        stage_state = state["stage"]

        ccd_color, slide_color = ("b", "k")  # chip slide (stopped)
        if (
            "error" in connection_state
            or "offline" in connection_state
            or "homed" not in stage_state
        ):
            slide_color = "r"
        elif "moving" in stage_state:
            slide_color = "y"

        ll = (0, 0)
        width = 2048
        height = 1024
        ccd = patches.Rectangle(ll, width, height, color=ccd_color, fill=False)
        slide_origin = 3500
        slide_corners = np.array(
            [
                ((-50, slide_origin + 50)),  # top left
                (-50, pos),  # bottom left
                (2048 + 50, pos),  # bottom right
                (2048 + 50, slide_origin + 50),  # top right
            ]
        )
        slide = patches.Polygon(
            slide_corners, closed=False, alpha=0.8, color=slide_color
        )
        self.ax.clear()
        self.ax.add_patch(ccd)
        self.ax.add_patch(slide)
        self.ax.set_xlim(-80, 2048 + 80)
        self.ax.set_ylim(-200, slide_origin + 100)
        self.ax.set_aspect("equal")
        self.ax.set_axis_off()
        self.canvas.draw()

    def setExpertLevel(self):
        """
        Modifies widget according to expertise level, which in this
        case is just matter of hiding or revealing the LED option
        and changing the lower limit on the exposure button.
        """
        g = get_root(self).globals
        level = g.cpars["expert_level"]
        if level == 0:
            self.reset.grid_forget()
            self.enable.grid_forget()
            self.disable.grid_forget()
            self.restore.grid_forget()
            self.stop.grid_forget()
        else:
            self.stop.grid(row=2, column=0)
            self.disable.grid(row=2, column=1)
            self.enable.grid(row=2, column=2)
            self.reset.grid(row=3, column=0)
            self.restore.grid(row=3, column=1)

    @inlineCallbacks
    def action(self, *comm):
        """
        Send a command to the focal plane slide
        """
        session = get_root(self).globals.session
        if not session:
            self.print_message("no session")
            return

        if comm[0] == "block":
            topic = "hipercam.slide.rpc.stage.move"
            self.set_slide_target_position(BLOCK_POS)
            # allow time for statemachines to step forward
            # this means we state is updated to out of position before moving
            yield async_sleep(SLEEP_BEFORE_MOVE)
        elif comm[0] == "unblock":
            topic = "hipercam.slide.rpc.stage.move"
            self.set_slide_target_position(UNBLOCK_POS)
            yield async_sleep(SLEEP_BEFORE_MOVE)
        elif comm[0] == "goto":
            topic = "hipercam.slide.rpc.stage.move"
            self.set_slide_target_position(int(comm[1]))
            yield async_sleep(SLEEP_BEFORE_MOVE)
        else:
            topic = f"hipercam.slide.rpc.{comm[0]}"

        try:
            yield session.call(topic)
        except SerializationError:
            pass
        except Exception as err:
            g = get_root(self).globals
            msg = err.error_message() if hasattr(err, "error_message") else str(err)
            g.clog.warn("error in slide command:" + msg)

    def set_slide_target_position(self, pos):
        topic = "hipercam.slide.target_position"
        session = get_root(self).globals.session
        if session:
            session.publish(topic, pos)

    def send_message(self, topic, msg):
        session = get_root(self).globals.session
        if session:
            session.publish(topic, msg)

    def print_message(self, msg):
        self.label.delete(1.0, tk.END)
        self.label.insert(tk.END, msg + "\n")

    def on_telemetry(self, package_data):
        telemetry = pickle.loads(package_data)

        try:
            pos = telemetry["position"]["current"].value
            targ = telemetry["position"]["target"].value
        except Exception:
            pos = telemetry["position"]["current"]
            targ = telemetry["position"]["target"]

        state = telemetry["state"]
        if "error" in state["connection"] or "offline" in state["connection"]:
            self.connect.config(
                command=lambda: self.action("connection.connect"), text="conn"
            )
        else:
            self.connect.config(
                command=lambda: self.action("connection.disconnect"), text="disconn"
            )

        str = f"{telemetry['timestamp']}:\n"
        status = "/".join(state["stage"][4:])
        str += f"pos: curr={pos:.0f}, targ={targ:.0f}\n{status}\n\n"
        self.update_mimic(telemetry)
        self.print_message(str)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# This sub-module is destined for common non-package specific utility
# functions that will ultimately be merged into `astropy.utils`

# As far as possible, utils contains classes of generic use, such as
# PosInt for positive integer input. See e.g. 'hcam' for more instrument
# dependent components.

from __future__ import print_function, unicode_literals, absolute_import, division
import six
import numpy as np
import pickle

from hcam_devices.devices.meerstetter import TEC_CURRENT_LIMIT
from hcam_widgets import widgets as w
from hcam_widgets.tkutils import get_root, addStyle
from .alarms import NoAlarmState

if not six.PY3:
    import Tkinter as tk
else:
    import tkinter as tk


class BoolFormatter(object):
    """
    A simple class to convert integer values to True/False strings.

    This class is a bit of a Kludge to make boolean values work nicely with the
    HardwareDisplayWidget, which was originally written with floating point numbers
    in mind. It provides one function, format, which converts 0 to 'ERROR' and 1 to
    'OK'.
    """
    def format(self, val):
        if val == 1:
            return 'OK'
        else:
            return 'ERROR'


class HardwareDisplayWidget(tk.Frame):
    """
    A widget that displays and checks the status of a piece of hardware.

    Consists of a label naming the piece of hardware, and an Ilabel to display the results of
    the hardware check.

    Has a telemetry topic and a callback to be used whenever a telemetry message is published
    to that topic. It is the responsibility of implementing widgets to subscribe to that
    topic with the correct callback.

    Each HardwareDisplayWidget has its own status (ok/nok) and an alarm state which can
    be NoAlarm, ActiveAlarm, AcknowledgedAlarm.

    Arguments
    ----------
    parent : tk.Widget
        parent widget
    kind : string
        type of hardware item, e.g. 'pressure'
    name : string
        name to display on label
    fmt : string
        format string for displaying results
    lower_limit : float, `~astropy.units.Quantity`
        lower limit for hardware check
    upper_limit : float, `~astropy.units.Quantity`
        upper limit for hardware check
    """
    def __init__(self, parent, topic, kind, name, lower_limit, upper_limit):
        tk.Frame.__init__(self, parent)
        self.parent = parent
        self.topic = topic
        self.name = name
        self.kind = kind
        self.ok = True
        self.fmt = '{:.1f}'
        self.upper_limit = upper_limit
        self.lower_limit = lower_limit
        self.set_state(NoAlarmState)
        tk.Label(self, text=self.name + ':', width=9, anchor=tk.E).pack(side=tk.LEFT, anchor=tk.N, padx=5)
        self.label = w.Ilabel(self, text='nan', width=9)
        self.label.pack(side=tk.LEFT, anchor=tk.N, padx=5)

    def on_telemetry(self, data):
        """
        Try and get the result of hardware update from the queue. If it's not ready yet,
        reschedule this check for 200ms later.
        """
        try:
            telemetry = pickle.loads(data)
            val, errmsg = self.process_telemetry(telemetry)
            g = get_root(self.parent).globals
            if errmsg is not None:
                g.clog.warn('Could not update {} for {}: {}'.format(self.kind, self.name, errmsg))

            self.label.configure(text=self.fmt.format(val), bg=g.COL['main'])

            if errmsg is None and val <= self.upper_limit and val >= self.lower_limit:
                self.ok = True
            elif not isinstance(val, str) and np.isnan(val) and errmsg is None:
                # no error and nan returned means checking disabled
                self.ok = True
            elif isinstance(val, str) and val == '??':
                # checking disabled
                self.ok = True
            else:
                self.ok = False

            if not self.ok:
                self.label.configure(bg=g.COL['warn'])
        except Exception as err:
            g = get_root(self.parent).globals
            g.clog.warn('could not process {} telemetry for {}: {}'.format(
                self.kind, self.name, str(err)
            ))

    def _extract_telemetry(self, attr, telemetry):
        if attr in telemetry:
            val = telemetry[attr]
            if hasattr(val, 'value'):
                val = val.value
            return val, None
        else:
            return np.nan, 'cannot parse telemetry'

    def process_telemetry(self):
        raise NotImplementedError('concrete class must implement process_telemetry')

    def acknowledge_alarm(self):
        self._state.acknowledge_alarm(self)

    def raise_alarm(self):
        self._state.raise_alarm(self)

    def cancel_alarm(self):
        self._state.cancel_alarm(self)

    def set_state(self, state):
        self._state = state


class MeerstetterWidget(HardwareDisplayWidget):
    """
    Get CCD info from Meerstetters
    """
    def __init__(self, parent, ccdnum, name, kind, lower_limit, upper_limit):
        meerstetter = 1 + ccdnum // 4
        topic = 'hipercam.meerstetter{}.telemetry'.format(meerstetter)
        HardwareDisplayWidget.__init__(self, parent, topic, kind, name,
                                       lower_limit, upper_limit)
        self.address = [1, 2, 3, 1, 2][ccdnum-1]
        if kind == 'peltier power':
            self.fmt = '{:.0f}'
        elif kind == 'status':
            self.fmt = '{:s}'

    def process_telemetry(self, telemetry):
        g = get_root(self.parent).globals
        if g.cpars['ccd_temp_monitoring_on']:
            if self.kind == 'status':
                return self._extract_telemetry('status{}'.format(self.address),
                                               telemetry)
            if self.kind == 'temperature':
                return self._extract_telemetry('temperature{}'.format(self.address),
                                               telemetry)
            elif self.kind == 'heatsink temperature':
                return self._extract_telemetry('heatsink{}'.format(self.address),
                                               telemetry)
            elif self.kind == 'peltier power':
                current, msg = self._extract_telemetry('current{}'.format(self.address),
                                                       telemetry)
                return 100 * current / TEC_CURRENT_LIMIT, None
            else:
                raise ValueError('unknown kind: {}'.format(self.kind))
        else:
            value = '??' if self.kind == 'status' else np.nan
            return value, None


class ChillerWidget(HardwareDisplayWidget):
    """
    Get Temperature from Chiller
    """
    def __init__(self, parent, lower_limit, upper_limit):
        topic = 'hipercam.chiller.telemetry'
        HardwareDisplayWidget.__init__(self, parent, topic, 'temperature', 'CHILLER',
                                       lower_limit, upper_limit)

    def process_telemetry(self, telemetry):
        g = get_root(self.parent).globals
        if g.cpars['chiller_temp_monitoring_on']:
            return self._extract_telemetry('temperature', telemetry)
        else:
            return np.nan, None


class RackSensorWidget(HardwareDisplayWidget):
    """
    Get Temperature and Humidity from Rack Sensor
    """
    def __init__(self, parent, lower_limit, upper_limit):
        topic = 'hipercam.gtc.telemetry'
        HardwareDisplayWidget.__init__(self, parent, topic, 'temperature', 'RACK',
                                       lower_limit, upper_limit)

    def process_telemetry(self, telemetry):
        g = get_root(self.parent).globals
        if g.cpars['chiller_temp_monitoring_on']:
            return self._extract_telemetry('rack_temp', telemetry)
        else:
            return np.nan, None


class FlowRateWidget(HardwareDisplayWidget):
    """
    Flow rates from honeywell
    """
    def __init__(self, parent, pen_number, name, lower_limit, upper_limit):
        topic = 'hipercam.flow.telemetry'
        HardwareDisplayWidget.__init__(self, parent, topic, 'flow rate', name,
                                       lower_limit, upper_limit)
        self.pen_number = pen_number
        self.fmt = '{:.2f}'

    def process_telemetry(self, telemetry):
        g = get_root(self.parent).globals
        if g.cpars['flow_monitoring_on']:
            attr = 'pen{}'.format(self.pen_number)
            return self._extract_telemetry(attr, telemetry)
        else:
            return np.nan, None


class VacuumWidget(HardwareDisplayWidget):
    """
    Vacuum pressure from gauge
    """
    def __init__(self, parent, ccdnum, name, lower_limit, upper_limit):
        topic = 'hipercam.pressure{}.telemetry'.format(ccdnum)
        HardwareDisplayWidget.__init__(self, parent, topic, 'pressure', name,
                                       lower_limit, upper_limit)
        self.fmt = '{:.2E}'

    def process_telemetry(self, telemetry):
        g = get_root(self.parent).globals
        if g.cpars['ccd_vac_monitoring_on']:
            val, msg = self._extract_telemetry('pressure', telemetry)
            return 1000*val, msg
        else:
            return np.nan, None


class FocalPlaneSlideWidget(HardwareDisplayWidget):
    """
    Focal plane Slide
    """
    def __init__(self, parent, name, lower_limit, upper_limit):
        topic = 'hipercam.slide.telemetry'
        HardwareDisplayWidget.__init__(self, parent, topic, 'position', name,
                                       lower_limit, upper_limit)
        self.fmt = '{:.0f}'

    def process_telemetry(self, telemetry):
        g = get_root(self).globals
        if g.cpars['focal_plane_slide_on']:
            position_dict, msg = self._extract_telemetry('position', telemetry)
            current = position_dict['current']
            if hasattr(current, 'value'):
                return position_dict['current'].value, msg
            else:
                return position_dict['current'], msg
        else:
            return np.nan, None


class CCDInfoWidget(tk.Toplevel):
    """
    A child window to monitor and show the status of the CCD heads.

    Keeps track of vacuum levels, CCD temperatures, flow rates etc. Normally this
    window is hidden, but can be revealed from the main GUIs menu. It will also
    appear automatically if there any issues, and should play a sound to notify
    users.
    """
    def __init__(self, parent):
        tk.Toplevel.__init__(self, parent)
        self.transient(parent)
        self.parent = parent

        addStyle(self)
        self.title("CCD Head Status")

        # do not display on creation
        self.withdraw()

        # dont destroy when we click the close button
        self.protocol('WM_DELETE_WINDOW', self.withdraw)

        # create label frames
        self.status_frm = tk.LabelFrame(self, text='Meerstetter status', padx=4, pady=4)
        self.temp_frm = tk.LabelFrame(self, text='Temperatures (C)', padx=4, pady=4)
        self.heatsink_frm = tk.LabelFrame(self, text='Heatsink Temps (C)', padx=4, pady=4)
        self.peltier_frm = tk.LabelFrame(self, text='Peltier Powers (%)', padx=4, pady=4)
        self.flow_frm = tk.LabelFrame(self, text='Flow Rates (l/min)', padx=4, pady=4)
        self.vac_frm = tk.LabelFrame(self, text='Vacuums (mbar)', padx=4, pady=4)

        # variables to store hardware widgets
        g = get_root(self).globals
        self.ms_status = []
        self.ccd_temps = []
        self.heatsink_temps = []
        self.peltier_powers = []
        self.ccd_flow_rates = []
        self.vacuums = []
        if g.cpars['telins_name'].lower() == 'wht':
            self.chiller_temp = ChillerWidget(self.temp_frm, g.cpars['chiller_temp_lower'],
                                              g.cpars['chiller_temp_upper'])
        else:
            self.chiller_temp = RackSensorWidget(self.temp_frm, g.cpars['rack_temp_lower'],
                                                 g.cpars['rack_temp_upper'])

        self.ngc_flow_rate = FlowRateWidget(self.flow_frm, 6, 'NGC',
                                            g.cpars['ngc_flow_lower'], g.cpars['ngc_flow_upper'])

        # populate CCD frames
        for iccd in range(5):
            ccdnum = iccd+1
            name = 'CCD {}'.format(iccd+1)
            # meerstetter widgets
            self.ms_status.append(
                MeerstetterWidget(self.status_frm, ccdnum, name, 'status', 'run', 'run')
            )
            self.ccd_temps.append(
                MeerstetterWidget(self.temp_frm, ccdnum, name, 'temperature',
                                  g.cpars['ccd_temp_lower'],
                                  g.cpars['ccd_temp_upper'])
            )
            self.heatsink_temps.append(
                MeerstetterWidget(self.heatsink_frm, ccdnum, name, 'heatsink temperature',
                                  g.cpars['ccd_sink_temp_lower'],
                                  g.cpars['ccd_sink_temp_upper'])
            )
            self.peltier_powers.append(
                MeerstetterWidget(self.peltier_frm, ccdnum, name, 'peltier power',
                                  g.cpars['ccd_peltier_lower'],
                                  g.cpars['ccd_peltier_upper'])
            )

            # grid
            self.ms_status[-1].grid(
                    row=int(iccd/3), column=iccd % 3, padx=5, sticky=tk.W
                )
            self.ccd_temps[-1].grid(
                    row=int(iccd/3), column=iccd % 3, padx=5, sticky=tk.W
                )
            self.heatsink_temps[-1].grid(
                row=int(iccd/3), column=iccd % 3, padx=5, sticky=tk.W
            )
            self.peltier_powers[-1].grid(
                row=int(iccd/3), column=iccd % 3, padx=5, sticky=tk.W
            )

            # flow rates
            pen_number = iccd+1
            self.ccd_flow_rates.append(
                FlowRateWidget(self.flow_frm, pen_number, name,
                               g.cpars['ccd_flow_lower'], g.cpars['ccd_flow_upper'])
            )
            self.ccd_flow_rates[-1].grid(
                    row=int(iccd/3), column=iccd % 3, padx=5, sticky=tk.W
            )

            # vacuum gauges
            name = 'CCD {}'.format(iccd+1)
            self.vacuums.append(
                VacuumWidget(self.vac_frm, iccd+1, name, -1e6, 5e-3)
            )
            self.vacuums[-1].grid(
                    row=int(iccd/3), column=iccd % 3, padx=5, sticky=tk.W
            )

        # hidden slide widget (not displayed, but used for JSON output)
        self.fpslide = FocalPlaneSlideWidget(self, 'slide', -100, 4000)

        # now for one-off items
        self.chiller_temp.grid(row=1, column=2, padx=5, sticky=tk.W)
        self.ngc_flow_rate.grid(row=1, column=2, padx=5, sticky=tk.W)

        # grid frames
        self.status_frm.grid(row=0, column=0, padx=4, pady=4, sticky=tk.W)
        self.temp_frm.grid(row=1, column=0, padx=4, pady=4, sticky=tk.W)
        self.heatsink_frm.grid(row=2, column=0, padx=4, pady=4, sticky=tk.W)
        self.peltier_frm.grid(row=3, column=0, padx=4, pady=4, sticky=tk.W)
        self.flow_frm.grid(row=4, column=0, padx=4, pady=4, sticky=tk.W)
        self.vac_frm.grid(row=5, column=0, padx=4, pady=4, sticky=tk.W)

        self.after_id = self.after(10000, self.raise_if_nok)

    def _getVal(self, widg):
        """
        Return value from widget if set, else return -99.
        """
        return -99 if widg.label['text'].lower() == 'nan' else float(widg.label['text'])

    def dumpJSON(self):
        """
        Encodes current hw data to JSON compatible dictionary
        """
        data = dict()
        for i in range(5):
            ccd = i+1
            data['ccd{}temp'.format(ccd)] = self._getVal(self.ccd_temps[i])
            data['ccd{}vac'.format(ccd)] = self._getVal(self.vacuums[i])
            data['ccd{}flow'.format(ccd)] = self._getVal(self.ccd_flow_rates[i])
        data['fpslide'] = self._getVal(self.fpslide)
        return data

    @property
    def ok(self):
        okl = [self.chiller_temp.ok, self.ngc_flow_rate.ok]
        okl += [ms_state.ok for ms_state in self.ms_status]
        okl += [ccd_temp.ok for ccd_temp in self.ccd_temps]
        okl += [vac.ok for vac in self.vacuums]
        okl += [flow.ok for flow in self.ccd_flow_rates]
        okl += [power.ok for power in self.peltier_powers]
        return all(okl)

    def raise_if_nok(self):
        widgets = [self.chiller_temp, self.ngc_flow_rate]
        widgets.extend(self.ccd_temps)
        widgets.extend(self.ms_status)
        widgets.extend(self.ccd_flow_rates)
        widgets.extend(self.peltier_powers)
        for widget in widgets:
            if not widget.ok:
                self.deiconify()
                widget.raise_alarm()

        self.after_id = self.after(10000, self.raise_if_nok)

    @property
    def telemetry_topics(self):
        widgets = [self.chiller_temp, self.ngc_flow_rate]
        widgets.extend(self.ccd_temps)
        widgets.extend(self.ms_status)
        widgets.extend(self.ccd_flow_rates)
        widgets.extend(self.peltier_powers)
        widgets.extend(self.heatsink_temps)
        widgets.extend(self.vacuums)
        widgets.append(self.fpslide)
        topics = [
            (widget.topic, widget.on_telemetry) for widget in widgets
        ]
        return topics



# Utility to query and set temps via Ethernet on Meerstetter
from __future__ import absolute_import, unicode_literals, print_function, division
import six
import pickle

# GUI imports
from hcam_widgets.widgets import RangedInt
from hcam_widgets.tkutils import get_root, addStyle

from twisted.internet.defer import inlineCallbacks

if not six.PY3:
    import Tkinter as tk
else:
    import tkinter as tk


class CCDTempFrame(tk.LabelFrame):
    """
    Self-contained widget to control CCD temps and reset TECS.
    """

    def __init__(self, master):
        tk.LabelFrame.__init__(self, master, text="CCD TECs", padx=10, pady=4)
        # Top for table of buttons
        top = tk.Frame(self)

        tk.Label(top, text="CCD1").grid(row=0, column=0)
        tk.Label(top, text="CCD2").grid(row=0, column=1)
        tk.Label(top, text="CCD3").grid(row=0, column=2)
        tk.Label(top, text="CCD4").grid(row=0, column=3)
        tk.Label(top, text="CCD5").grid(row=0, column=4)

        # maps ccd number to meerstetter, meerstetter address
        self.mapping = {1: (1, 1), 2: (1, 2), 3: (1, 3), 4: (2, 1), 5: (2, 2)}

        self.temp_entry_widgets = {}
        self.setpoint_displays = {}
        self.reset_buttons = {}
        width = 8
        for i in range(1, 6):
            ival = 5
            self.temp_entry_widgets[i] = RangedInt(
                top, ival, -100, 20, None, True, width=width
            )
            self.temp_entry_widgets[i].grid(row=1, column=i - 1)
            self.setpoint_displays[i] = tk.Label(top, text="nan", width=width)
            self.setpoint_displays[i].grid(row=2, column=i - 1)
            self.reset_buttons[i] = tk.Button(
                top,
                fg="black",
                width=width,
                text="Reset",
                command=lambda ccd=i: self.reset(ccd),
            )
            self.reset_buttons[i].grid(row=3, column=i - 1)

        # bind enter to set value routine
        for i in range(1, 6):
            widget = self.temp_entry_widgets[i]
            widget.unbind("<Return>")
            widget.bind("<Return>", lambda event, ccd=i: self.update(ccd))

        top.pack(pady=2)
        addStyle(self)

        self.telemetry_topics = [
            (
                "hipercam.meerstetter1.telemetry",
                lambda data: self.on_telemetry(1, data),
            ),
            (
                "hipercam.meerstetter2.telemetry",
                lambda data: self.on_telemetry(2, data),
            ),
        ]

    def on_telemetry(self, ms, data):
        g = get_root(self).globals
        try:
            telemetry = pickle.loads(data)
        except Exception as err:
            wrn = "cannot parse telemetry from Meerstetter {}: {}"
            g.clog.warn(wrn.format(ms, str(err)))
        else:
            if ms == 1:
                widget_numbers = (1, 2, 3)
            else:
                widget_numbers = (4, 5)
            for i, num in enumerate(widget_numbers):
                key = "target_temperature{}".format(i + 1)
                setpoint = telemetry[key].value
                widget = self.setpoint_displays[num]
                widget.configure(text=str(setpoint))

    @inlineCallbacks
    def update(self, ccd):
        g = get_root(self).globals
        if not g.cpars["ccd_temp_monitoring_on"]:
            g.clog.warn(
                "Temperature monitoring disabled. Will not update CCD{}".format(ccd)
            )
            return
        g.clog.info("Updating CCD{}".format(ccd))
        widget = self.temp_entry_widgets[ccd]
        val = widget.value()
        g.clog.info("desired setpoint " + str(val))

        try:
            session = get_root(self).globals.session
            topic = "hipercam.ccd{}.setpoint".format(ccd)
            g.clog.info("publishing to topic " + topic)
            yield session.publish(topic, float(val))
        except Exception:
            g.clog.warn("Unable to update setpoint for CCD{}".format(ccd))

    @inlineCallbacks
    def reset(self, ccd):
        g = get_root(self).globals
        if not g.cpars["ccd_temp_monitoring_on"]:
            g.clog.warn(
                "Temperature monitoring disabled. Will not reset CCD{}".format(ccd)
            )
            return
        g.clog.info("Resetting TEC on CCD{}".format(ccd))
        ms, address = self.mapping[ccd]
        try:
            rpc = "hipercam.meerstetter{}.rpc.reset".format(ms)
            session = get_root(self).globals.session
            yield session.call(rpc, address)
        except Exception as err:
            g.clog.warn("Unable to reset TEC {}".format(ccd))
            msg = err.error_message() if hasattr(err, "error_message") else str(err)
            g.clog.warn(msg)



# Some utility tools to handle rising and setting etc.
from __future__ import print_function, unicode_literals, absolute_import, division
import warnings
import numpy as np
from numpy.polynomial.polynomial import polyval

# astropy
from astropy import coordinates as coord
from astropy.time import Time
from astropy import units as u

MAGIC_TIME = Time(-999, format='jd')


def _equation_of_time(t):
    """
    Find the difference between apparent and mean solar time

    Parameters
    ----------
    t : `~astropy.time.Time`
        times (array)

    Returns
    ----------
    ret1 : `~astropy.units.Quantity`
        the equation of time
    """

    # Julian centuries since J2000.0
    T = (t - Time("J2000")).to(u.year).value / 100

    # obliquity of ecliptic (Meeus 1998, eq 22.2)
    poly_pars = (84381.448, 46.8150, 0.00059, 0.001813)
    eps = u.Quantity(polyval(T, poly_pars), u.arcsec)
    y = np.tan(eps/2)**2

    # Sun's mean longitude (Meeus 1998, eq 25.2)
    poly_pars = (280.46646, 36000.76983, 0.0003032)
    L0 = u.Quantity(polyval(T, poly_pars), u.deg)

    # Sun's mean anomaly (Meeus 1998, eq 25.3)
    poly_pars = (357.52911, 35999.05029, 0.0001537)
    M = u.Quantity(polyval(T, poly_pars), u.deg)

    # eccentricity of Earth's orbit (Meeus 1998, eq 25.4)
    poly_pars = (0.016708634, -0.000042037, -0.0000001267)
    e = polyval(T, poly_pars)

    # equation of time, radians (Meeus 1998, eq 28.3)
    eot = (y * np.sin(2*L0) - 2*e*np.sin(M) + 4*e*y*np.sin(M)*np.cos(2*L0) -
           0.5*y**2 * np.sin(4*L0) - 5*e**2 * np.sin(2*M)/4) * u.rad
    return eot.to(u.hourangle)


def _astropy_time_from_LST(t, LST, location, prev_next):
    """
    Convert a Local Sidereal Time to an astropy Time object.

    The local time is related to the LST through the RA of the Sun.
    This routine uses this relationship to convert a LST to an astropy
    time object.

    Returns
    -------
    ret1 : `~astropy.time.Time`
        time corresponding to LST
    """
    # now we need to figure out time to return from LST
    raSun = coord.get_sun(t).ra

    # calculate Greenwich Apparent Solar Time, which we will use as ~UTC for now
    lon = location.lon
    solarTime = LST - raSun + 12*u.hourangle - lon

    # assume this is on the same day as supplied time, and fix later
    first_guess = Time(
        u.d*int(t.mjd) + u.hour*solarTime.wrap_at('360d').hour,
        format='mjd'
    )

    # Equation of time is difference between GAST and UTC
    eot = _equation_of_time(first_guess)
    first_guess = first_guess - u.hour * eot.value

    if prev_next == 'next':
        # if 'next', we want time to be greater than given time
        mask = first_guess < t
        rise_set_time = first_guess + mask * u.sday
    else:
        # if 'previous', we want time to be less than given time
        mask = first_guess > t
        rise_set_time = first_guess - mask * u.sday
    return rise_set_time


def _rise_set_trig(t, target, location, prev_next, rise_set):
    """
    Crude time at next rise/set of ``target`` using spherical trig.

    This method is ~15 times faster than `_calcriseset`,
    and inherently does *not* take the atmosphere into account.

    The time returned should not be used in calculations; the purpose
    of this routine is to supply a guess to `_calcriseset`.

    Parameters
    ----------
    t : `~astropy.time.Time` or other (see below)
        Time of observation. This will be passed in as the first argument to
        the `~astropy.time.Time` initializer, so it can be anything that
        `~astropy.time.Time` will accept (including a `~astropy.time.Time`
        object)

    target : `~astropy.coordinates.SkyCoord`
        Position of target or multiple positions of that target
        at multiple times (if target moves, like the Sun)

    location : `~astropy.coordinates.EarthLocation`
        Observatory location

    prev_next : str - either 'previous' or 'next'
        Test next rise/set or previous rise/set

    rise_set : str - either 'rising' or 'setting'
        Compute prev/next rise or prev/next set

    Returns
    -------
    ret1 : `~astropy.time.Time`
        Time of rise/set
    """
    dec = target.transform_to(coord.ICRS).dec
    lat = location.lat
    cosHA = -np.tan(dec)*np.tan(lat.radian)
    # find the absolute value of the hour Angle
    HA = coord.Longitude(np.fabs(np.arccos(cosHA)))
    # if rise, HA is -ve and vice versa
    if rise_set == 'rising':
        HA = -HA
    # LST = HA + RA
    LST = HA + target.ra

    return _astropy_time_from_LST(t, LST, location, prev_next)


def calc_riseset(t, target_name, location, prev_next, rise_set, horizon):
    """
    Time at next rise/set of ``target``.

    Parameters
    ----------
    t : `~astropy.time.Time` or other (see below)
        Time of observation. This will be passed in as the first argument to
        the `~astropy.time.Time` initializer, so it can be anything that
        `~astropy.time.Time` will accept (including a `~astropy.time.Time`
        object)

    target_name : str
        'moon' or 'sun'

    location : `~astropy.coordinates.EarthLocation`
        Observatory location

    prev_next : str - either 'previous' or 'next'
        Test next rise/set or previous rise/set

    rise_set : str - either 'rising' or 'setting'
        Compute prev/next rise or prev/next set

    location : `~astropy.coordinates.EarthLocation`
        Location of observer

    horizon : `~astropy.units.Quantity`
        Degrees above/below actual horizon to use
        for calculating rise/set times (i.e.,
        -6 deg horizon = civil twilight, etc.)

    Returns
    -------
    ret1 : `~astropy.time.Time`
        Time of rise/set
    """
    target = coord.get_body(target_name, t)
    t0 = _rise_set_trig(t, target, location, prev_next, rise_set)
    grid = t0 + np.linspace(-4*u.hour, 4*u.hour, 10)
    altaz_frame = coord.AltAz(obstime=grid, location=location)
    target = coord.get_body(target_name, grid)
    altaz = target.transform_to(altaz_frame)
    time_limits, altitude_limits = _horiz_cross(altaz.obstime, altaz.alt,
                                                rise_set, horizon)
    return _two_point_interp(time_limits, altitude_limits, horizon)


@u.quantity_input(horizon=u.deg)
def _horiz_cross(t, alt, rise_set, horizon=0*u.degree):
    """
    Find time ``t`` when values in array ``a`` go from
    negative to positive or positive to negative (exclude endpoints)

    ``return_limits`` will return nearest times to zero-crossing.

    Parameters
    ----------
    t : `~astropy.time.Time`
        Grid of times
    alt : `~astropy.units.Quantity`
        Grid of altitudes
    rise_set : {"rising",  "setting"}
        Calculate either rising or setting across the horizon
    horizon : float
        Number of degrees above/below actual horizon to use
        for calculating rise/set times (i.e.,
        -6 deg horizon = civil twilight, etc.)

    Returns
    -------
    Returns the lower and upper limits on the time and altitudes
    of the horizon crossing.
    """
    if rise_set == 'rising':
        # Find index where altitude goes from below to above horizon
        condition = (alt[:-1] < horizon) * (alt[1:] > horizon)
    elif rise_set == 'setting':
        # Find index where altitude goes from above to below horizon
        condition = (alt[:-1] > horizon) * (alt[1:] < horizon)

    if np.count_nonzero(condition) == 0:
        warnmsg = ('Target does not cross horizon={} within '
                   '8 hours of trigonometric estimate'.format(horizon))
        warnings.warn(warnmsg)

        # Fill in missing time with MAGIC_TIME
        time_inds = np.nan
        times = [np.nan, np.nan]
        altitudes = [np.nan, np.nan]
    else:
        time_inds = np.nonzero(condition)[0][0]
        times = t[time_inds:time_inds+2]
        altitudes = alt[time_inds:time_inds+2]

    return times, altitudes


@u.quantity_input(horizon=u.deg)
def _two_point_interp(times, altitudes, horizon=0*u.deg):
    """
    Do linear interpolation between two ``altitudes`` at
    two ``times`` to determine the time where the altitude
    goes through zero.

    Parameters
    ----------
    times : `~astropy.time.Time`
        Two times for linear interpolation between

    altitudes : array of `~astropy.units.Quantity`
        Two altitudes for linear interpolation between

    horizon : `~astropy.units.Quantity`
        Solve for the time when the altitude is equal to
        reference_alt.

    Returns
    -------
    t : `~astropy.time.Time`
        Time when target crosses the horizon

    """
    if not isinstance(times, Time):
        return MAGIC_TIME
    else:
        slope = (altitudes[1] - altitudes[0])/(times[1].jd - times[0].jd)
        return Time(times[1].jd - ((altitudes[1] - horizon)/slope).value,
                    format='jd')
